/** @file
 * This file is intended to be executed by a WCCOActrl manager. It initializes
 *  and runs the core script.
 */
#uses "fwRPC/fwRPC_core.ctl"

/**
 * Representation of user arguments passed to the manager.
 * nullptr represents a missing argument
 */
struct FwRPC_Args
{
  shared_ptr<int> startingWorkers = nullptr;
  shared_ptr<int> maxWorkers = nullptr;
  shared_ptr<uint> minClientContextPoolSize = nullptr;
  shared_ptr<bool> dist = nullptr;
};


/**
 * Start the core component. If startingWorkers and maxWorkers is set to -1,
 * workers will not be automatically started/stopped.
 * 
 * Arguments passed to this manager can be the following: [-startingWorkers <int>] [-maxWorkers <int>] [-dist <bool>]
 *
 * @param startingWorkers (IN) number of workers to initialize on startup
 * @param maxWorkers      (IN) maximum number of workers that should be
 *                                running
 * @param dist            (IN) start in distributed mode
 * @param clientName      (IN) name of the core if in distributed mode
 */
int main(...) 
{
  // Here we handle the manager arguments
  dyn_string userArgsRaw;
  va_list managerArgs;
  
  const int len = va_start(managerArgs);
  for (int i = 1; i <= len; i++) {
    dynAppend(userArgsRaw, managerArgs[i]);
  }
  va_end(managerArgs);
  
  
  const FwRPC_Args originalUserArgs = parseUserArguments(userArgsRaw, startingWorkers, maxWorkers, dist);
  const FwRPC_Args userArgs = setMissingArgumetsToDefaultValues(originalUserArgs);
  
  const string clientName = "fwRPC_core_"+strrtrim(getSystemName(),":");
  

  int minClientContextPoolSize = -1;
  if(userArgs.minClientContextPoolSize != nullptr)
  {
    minClientContextPoolSize = (int) userArgs.minClientContextPoolSize;
  }

  fwRPC_core_initialize(userArgs.startingWorkers,
                        userArgs.maxWorkers,
                        minClientContextPoolSize,
                        userArgs.dist,
                        clientName);
  return 0;
}


/** 
 * This function parses user arguments.
 * If an argument was not explicitly defined, the class field will have nullptr value.
 * 
 * @see setMissingArgumetsToDefaultValues() to get the default values
 */
private FwRPC_Args parseUserArguments(const dyn_string userArgsRaw)
{
  FwRPC_Args userArgs;

  const string STARTING_WORKERS = "-startingWorkers";
  const string MAX_WORKERS = "-maxWorkers";
  const string MIN_CLIENT_CONTEXT_POOL_SIZE = "-minClientContextPoolSize";
  const string DIST = "-dist";
  
  bool allOKFlag = true;
  const int argsLen = dynlen(userArgsRaw);
  for(int i = 1; i <= argsLen; i++)
  {
    const string arg = userArgsRaw[i];
    if(arg == STARTING_WORKERS)
    {
      if(i < argsLen)
      {
        i++;
        userArgs.startingWorkers = new int(userArgsRaw[i]);
      }
      else
      {
        allOKFlag = false;
        DebugTN(__FILE__+":"+__FUNCTION__+">>> Cannot parse user arguments. Number of starting Workers expected", userArgsRaw);
      }
    }
    else if(arg == MAX_WORKERS)
    {
      if(i < argsLen)
      {
        i++;
        userArgs.maxWorkers = new int(userArgsRaw[i]);
      }
      else
      {
        allOKFlag = false;
        DebugTN(__FILE__+":"+__FUNCTION__+">>> Cannot parse user arguments. Number of starting Workers expected", userArgsRaw);
      }
    }
    else if(arg == MIN_CLIENT_CONTEXT_POOL_SIZE)
    {
      if(i < argsLen)
      {
        i++;
        userArgs.minClientContextPoolSize = new uint(userArgsRaw[i]);
      }
      else
      {
        allOKFlag = false;
        DebugTN(__FILE__+":"+__FUNCTION__+">>> Cannot parse user arguments. Number of minimum client context pool size expected", userArgsRaw);
      }
    }
    else if(arg == DIST)
    {
      if(i < argsLen)
      {
        i++;
        userArgs.dist = new bool(userArgsRaw[i]);
      }
      else
      {
        allOKFlag = false;
        DebugTN(__FILE__+":"+__FUNCTION__+">>> Cannot parse user arguments. Number of starting Workers expected", userArgsRaw);
      }
    }
    else
    {
      allOKFlag = false;
      DebugTN(__FILE__+":"+__FUNCTION__+">>> Argument at the following index was not recgnized:"+i, userArgsRaw);
    }
  }

  return userArgs;
}

/**
 * Replaces fields with nullptr values with default values.
 */
private FwRPC_Args setMissingArgumetsToDefaultValues(FwRPC_Args userArgs)
{
  if(userArgs.startingWorkers == nullptr)
  {
    userArgs.startingWorkers = new int(1);
  }
  if(userArgs.maxWorkers == nullptr)
  {
    userArgs.maxWorkers = new int(1);
  }
  if(userArgs.dist == nullptr)
  {
    userArgs.dist = new bool(true);
  }

  return userArgs;
}