/** @file
 *
 * This library contains commonly used constants and helper functions.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_UUID.ctl"
#uses "fwConfigs/fwDPELock.ctl"

#uses "CtrlSerialization"


/** 
 * @ingroup Constants
 * @{
 */

/**
 * DP/DPE consts
 * @{
 */
const string fwRPC_RBMBX_DPT            = "_FwRPC_RB_MBX";
const string fwRPC_DPE_VALUE            = ":_original.._value";
const string fwRPC_DPE_NO_VALUE_STORAGE = ":_original.._last_value_storage_off";
const string fwRPC_DP_SEPARATOR         = "/";
/* @} */


/**
 * Valid message types
 * @{
 */
const string fwRPC_DISTMSG_TYPE_REQUEST     = "REQ";
const string fwRPC_DISTMSG_TYPE_GET_RESULT  = "GRES";
const string fwRPC_DISTMSG_TYPE_SET_RESULT  = "SRES";
const string fwRPC_DISTMSG_TYPE_STATUS      = "STAT";
const string fwRPC_DISTMSG_TYPE_RECOVER     = "RCVR";

const string fwRPC_DISTMSG_TYPE_REGISTER    = "REG";
const string fwRPC_DISTMSG_TYPE_ACKNOWLEDGE = "ACK";
const string fwRPC_DISTMSG_TYPE_CANCEL      = "CNCL";
const string fwRPC_DISTMSG_TYPE_LISTSVC     = "SVC";
const string fwRPC_DISTMSG_TYPE_CONNECT     = "ON";
const string fwRPC_DISTMSG_TYPE_DISCONNECT  = "OFF";

const string fwRPC_DISTMSG_TYPE_WORKER_HEARTBEAT = "W_HB";
const string fwRPC_DISTMSG_TYPE_WORKER_STATE     = "W_ST";
/* @} */


/**
 * Request message parameters
 * @{
 */
const int fwRPC_DISTMSG_REQUEST_HDL     = 1;
const int fwRPC_DISTMSG_REQUEST_FUNC    = 2;
const int fwRPC_DISTMSG_REQUEST_PARAMS  = 3;
const int fwRPC_DISTMSG_REQUEST_TIMEOUT = 4;
/* @} */


/**
 * Result message parameters
 * @{
 */
const int fwRPC_DISTMSG_RESULT_HDL    = 1;
const int fwRPC_DISTMSG_RESULT_STATUS = 2;
const int fwRPC_DISTMSG_RESULT_OUTPUT = 3;
/* @} */


/**
 * Status message parameters
 *  @{
 */
const int fwRPC_DISTMSG_STATUS_HDL      = 1;
const int fwRPC_DISTMSG_STATUS_STATUS   = 2;
const int fwRPC_DISTMSG_STATUS_CALLBACK = 3;
/* @} */


/**
 * Recover message parameters
 * @{
 */
const int fwRPC_DISTMSG_RECOVER_HDL      = 1;
const int fwRPC_DISTMSG_RECOVER_CALLBACK = 2;
/* @} */


/**
 * Worker state message parameters
 * @{
 */
const int fwRPC_DISTMSG_WORKER_STATE_ID     = 1;
const int fwRPC_DISTMSG_WORKER_STATE_STATUS = 2;
const int fwRPC_DISTMSG_WORKER_STATE_JOBS   = 3;
const int fwRPC_DISTMSG_WORKER_STATE_LOCKED = 4;
/* @} */


/**
 * Possible client types
 * @{
 */
const string fwRPC_DIST_TAG_CLIENT = "clnt";
const string fwRPC_DIST_TAG_CORE   = "core";
const string fwRPC_DIST_TAG_WORKER = "work";
/* @} */


/**
 * Possible locality types
 * @{
 */
const string fwRPC_DIST_TAG_LOCAL  = "loc";
const string fwRPC_DIST_TAG_REMOTE = "rem";
/* @} */


/**
 * Endpoint node information
 * @{
 */
const int fwRPC_DIST_NODE_PROJECT = 1;
const int fwRPC_DIST_NODE_NAME    = 2;
/* @} */


/**
 * Messages targeted towards brokers should have this as dest. info.
 * @{
 */
const string fwRPC_DIST_BRKR = "brokermodule"; 
/* @} */


//------------------------------------------------------------------------------


/**
 * build DP name from DPT and model ID. Does not check if the DPT exists.
 *
 * @param dpt (IN, string) DPT
 * @param id  (IN, string) model id
 *
 * @returns resulting DP name
 *
 * @throws nothing
 */
string fwRPC_common_makeDpName(string dpt, string id) {
  return dpt + fwRPC_DP_SEPARATOR + id;
}


/**
 * Check if the given variable is a valid rpc handle or not.
 *
 * @param rpcHandle (IN, anytype) variable to check
 *
 * @return  true if rpcHandle is valid
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_common_isValidHdl(anytype rpcHandle) {
  if (getType(rpcHandle) != STRING_VAR) return false;
  if (strlen(rpcHandle) == 0) return false;
  if (!fwRPC_UUID_isValidUUID4(rpcHandle)) return false;

  return true;
}


/**
 * Unpack a packed message string.
 *
 * @param message       (IN,  string) message to unpack
 * @param id            (OUT, string) message id
 * @param type          (OUT, string) message type
 * @param targetProject (OUT, string) message targetProject
 * @param targetName    (OUT, string) message targetName
 * @param buffer        (OUT, string) message buffer
 *
 * @returns  0 on success
 *          -1 otherwise
 *
 * @throws nothing
 */
int fwRPC_common_unpackMessage(
  string message,
  string &id,
  string &type,
  string &targetProject,
  string &targetName,
  string &buffer
) {
  int start = 0;
  int end;

  end = strpos(message, " ", start);
  if (end < 0) return -1;
  id = substr(message, start, end);
  start = end+1;
  end = strpos(message, " ", start);
  if (end < 0) return -1;

  type = substr(message, start, end-start);
  start = end+1;
  end = strpos(message, " ", start);
  if (end < 0) return -1;

  targetProject = substr(message, start, end-start);
  start = end+1;
  end = strpos(message, " ", start);
  if (end < 0) return -1;

  targetName = substr(message, start, end-start);
  buffer = substr(message, end+1);

  return 0;
}


/**
 * Clean a project name, removing the final ':' char if present.
 *
 * @param name (IN, string) project name to clean
 *
 * @returns clean project name
 *
 * @throws nothing
 */
string fwRPC_common_cleanProjectName(string name) {
  if ((string) name[strlen(name)-1] != ":") return name;
  return substr(name, 0, strlen(name) -1);
}


/**
 * Check if a given name belongs to a client.
 *
 * @param name (IN, string) name to check
 *
 * @returns true if name represents client
 *          false otherwise
 * 
 * @throws nothing
 */
bool fwRPC_dist_isClient(string name) {
  return fwRPC_dist_nameMatches(name, fwRPC_DIST_TAG_CLIENT);
}


/**
 * Check if a given name belongs to a core.
 *
 * @param name (IN, string) name to check
 *
 * @returns true if name represents core
 *          false otherwise
 * 
 * @throws nothing
 */
bool fwRPC_dist_isCore(string name) {
  return fwRPC_dist_nameMatches(name, fwRPC_DIST_TAG_CORE);
}


/**
 * Check if a given name belongs to a worker.
 *
 * @param name (IN, string) name to check
 *
 * @returns true if name represents worker
 *          false otherwise
 * 
 * @throws nothing
 */
bool fwRPC_dist_isWorker(string name) {
  return fwRPC_dist_nameMatches(name, fwRPC_DIST_TAG_WORKER);
}


/**
 * Check if a given name belongs to a broker.
 *
 * @param name (IN) name to check
 *
 * @returns true if name represents broker
 *          false otherwise
 * 
 * @throws nothing
 */
bool fwRPC_dist_isBroker(string name) {
  return name == fwRPC_DIST_BRKR;
}


/**
 * Check if a given name matches against a given tag.
 *
 * @param name (IN, string) name to check
 * @param tag  (IN, string) tag to check
 *
 * @returns true  if name coincides with the tag
 *          false otherwise
 * 
 * @throws nothing
 */
bool fwRPC_dist_nameMatches(string name, string tag) {
  return substr(name, 4, 4) == tag;
}

/**
 * Strip the name string to just the unique identifier.
 *
 * @param name (IN) name to strip
 *
 * @returns the unique identifier
 *
 * @throws nothing
 */
string fwRPC_dist_getIdFromName(string name) {
  return substr(name, 9);
}


string fwRPC_dist_buildClientName(bool local, string typeTag, string id) {
  return (local ? fwRPC_DIST_TAG_LOCAL : fwRPC_DIST_TAG_REMOTE) + "_" +
         typeTag + "_" + id;
}


/**
 * Build a message string to be sent by a client or broker module.
 * Can be used for request or response messages.
 *
 * Type must be one of the defined fwRPC_DISTMSG_TYPE_*
 *
 * @param type    (IN, string)  message type
 * @param project (IN, string)  message project
 * @param name    (IN, string)  message name
 * @param payload (IN, anytype) payload
 *
 * @returns formatted message string
 *
 * @throws nothing
 */
string fwRPC_dist_buildMessage(
  string type,
  string project,
  string name,
  anytype payload
) {
  string msg = "";
  string serialized = "";
  string uid = fwRPC_UUID_generateUUID4();

  if (payload != "") serializeToXml(payload, serialized);
  strreplace(serialized, "\n", "");

  msg = uid      + " " +
        type     + " " +
        project  + " " + 
        name     + " " +
        serialized;

  return msg;
}


const uint fwRPC_DP_LOCK_TIMEOUT_sec = 5;
const uint fwRPC_DP_LOCK_LIFETIME_sec = 7;

/**
 * Locks the original value of the given DPE and prints out erros.
 * 
 * @param string dpe - DPE to be locked
 * @param int lockTimeout_sec = fwRPC_DP_LOCK_DELAY_sec - time (in seconds) for which the function will wait for the lock.
 *                  -- 0  - do not wait for lock at all (will exit with failure if the DPE is already locked)
 *                  -- <0 - wait with no limit until the lock is obtained.
 * @param int lockLifetime_sec = fwRPC_DP_LOCK_LIFETIME_sec - the maximum time for which the lock will be kept. If the lock
 *                  is not released within the lockLifeTime, then it will be force-unlocked by the 
 *                  lock manager. Specifying the value of zero will disable the lock-monitoring mechanism for this call.
 * @return true if the DPE original value was locked, else false
 */
bool fwRPC_common_dpLock(const string dpe,
                         const int lockTimeout_sec = fwRPC_DP_LOCK_TIMEOUT_sec,
                         const int lockLifetime_sec = fwRPC_DP_LOCK_LIFETIME_sec)
{
  dyn_string exceptionInfo;
  const bool locked = fwDPELock_tryLock(dpe, exceptionInfo, lockTimeout_sec, lockLifetime_sec);

  if(locked == false || dynlen(exceptionInfo) > 0)
  {
      DebugTN(__FUNCTION__, "fwRPC cannot lock the following DPE:", dpe,
                            "exceptionInfo", exceptionInfo, getStackTrace());
  }

  return locked;
}


/**
 * Unlock the original value of the given DPE.
 * 
 * @param string dpe - DPE to be unlocked
 * @return true of the DPE original value was succesfully unlocked, else false
 */
bool fwRPC_common_dpUnlock(const string dpe)
{
  dyn_string exceptionInfo;

  fwDPELock_unlock(dpe, exceptionInfo);
  if(dynlen(exceptionInfo) > 0)
  {
    DebugTN(__FUNCTION__, "fwRPC cannot unlock the following DPE:", dpe, "exceptionInfo", exceptionInfo, getStackTrace());
    return false; 
  }

  return true;
}

const string fwRPC_CONTEXT_POOL_DPT = "_FwRPC_ContextPool";
private const string fwRPC_POOL_CONTEXT_NAME_PREFIX = "_FwRPC_poolContext_";

/**
 * Ensures that the given pool of contexts has a certian minimal size.
 * 
 * @param uint minPoolSize - min size of the context pool
 * @param string poolName - name of the context pool
 * 
 * @return true if the operation was successful, else false
 */
bool fwRPC_common_ensureMinimumPoolSize(const uint minPoolSize, const string poolName)
{
  const string poolDp = fwRPC_common_getDpNameForContextPool(poolName);
  if(dpExists(poolDp) == false)
  {
    const int errFlag = dpCreate(poolDp, fwRPC_CONTEXT_POOL_DPT);
    const dyn_errClass lastErr = getLastError();
    if(errFlag < 0 || dynlen(lastErr) > 0)
    {
      DebugTN(__FUNCTION__, "A dpCreate() problem occured when ensuring the min fwRPC pool size", errFlag, lastErr);
      return false;
    }
  }

  const string availableContextsDpe = poolDp+".availableContexts";

  const bool isLocked = fwRPC_common_dpLock(availableContextsDpe);
  if(isLocked == true)
  {
    dyn_string contextNames;
    dpGet(availableContextsDpe, contextNames);

    const string contextNamePrefix = fwRPC_POOL_CONTEXT_NAME_PREFIX+poolName+"_";
    for(int i = 0; dynlen(contextNames) < minPoolSize; i++)
    {
      const string contextName = contextNamePrefix+(string)i;
      if(dynContains(contextNames, contextName) < 1)
      {
        dynAppend(contextNames, contextName);
      }
    }

    const int errFlag = dpSetWait(availableContextsDpe, contextNames);
    const dyn_errClass lastErr = getLastError();
    if(errFlag < 0 || dynlen(lastErr) > 0)
    {
      DebugTN(__FUNCTION__, "A problem with dpSet() when ensuring the min context pool size", availableContextsDpe, errFlag, lastErr);
      fwRPC_common_dpUnlock(availableContextsDpe);
      return false;
    }

    fwRPC_common_dpUnlock(availableContextsDpe);

    return true;
  }
  else
  {
    return false;
  }
}

/**
 * Returns DP name for given context pool
 * 
 * @param string poolName - name of a context pool we want to create DP name for
 * @return DP name of the context pool
 */
string fwRPC_common_getDpNameForContextPool(const string poolName)
{
  return "_FwRPC_pool_"+poolName;
}


/**
 * @return time converted to seconds since the beginning of the Epoch
 */
public long fwRPC_common_timeToSeconds(const time &timeToConvert)
{
  return (long) timeToConvert;
}


/**
 * @return time converted to milliseconds since the beginning of the Epoch
 */
public long fwRPC_common_timeToMilliSeconds(const time &timeToConvert)
{
  return (fwRPC_common_timeToSeconds(timeToConvert)*1000) + (long)milliSecond(timeToConvert);
}


/**
 * @return current time converted to milliseconds since the beginning of the Epoch
 */
public long fwRPC_common_timeGetCurrentTimeInMilliSeconds()
{
  return fwRPC_common_timeToMilliSeconds(getCurrentTime());
}