/** @file
 * This library contains the set of functions that handles the RPC core,
 *  responsible for initializing worker management and request
 *  distribution/dispatch modules.
 * Part of the server-side RPC library.
 *
 * Should be ran in it's own CTRL Manager. Only one instance should be running
 *  at any given time.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */


#uses "fwRPC/fwRPC_workerManager.ctl"
#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_workerModel.ctl"
#uses "fwRPC/fwRPC_requestModel.ctl"
#uses "fwRPC/fwRPC_functionModel.ctl"
#uses "fwRPC/fwRPC_functionManager.ctl"
#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_msgClient.ctl"
#uses "fwRPC/fwRPC_msgBroker.ctl"
#uses "fwRPC/fwRPC_debug.ctl"
#uses "fwRPC/fwRPC_client.ctl"


/**
 * Error codes
 */

private global const int fwRPC_CORE_E_SUCCESS = 0;
private global const int fwRPC_CORE_E_ERROR = -1;
private global const int fwRPC_CORE_E_REQ_NOT_FOUND = -2;

/**
 * End error codes
 */


/**
 * Request queue. Requests are enqueued here when a suitable worker has not been
 *  found yet.
 *
 * Structure:
 *  [x]: request model (see fwRPC_requestModel.ctl)
 */
private global dyn_dyn_mixed g_fwRPC_core_requestQueue;

/**
 * Mapping of requests currently in progress.
 * Structure:
 *  key   = requesst handle
 *  value = request object (see fwRPC_requestModel.ctl)
 */
private global mapping g_fwRPC_core_requests;

/**
 * The hdl of the last queued request since a restart was requested. This is to
 * protect against requests failing when made immediately before a restart was
 * initiated and their function definitions change.
 *
 * An empty string indicates that no legacy requests are waiting and we can
 * restart the workers.
 */
private global mixed g_fwRPC_core_last_queued_since_restart;

/**
 * Whether the core is running in distributed or local mode.
 * If distributed mode is on, then the core will communicate with fwRPC systems
 * running in other WCCOA systems.
 */
private global bool g_fwRPC_core_distributed;

/**
 * Default msgClient name
 */
private global const string _fwRPC_CORE_DEFAULT_NAME = "fwRPC_core";

/**
 * Defines whether or not the core is pending a function definition restart. If
 * this value is true, then all new requests will be enqueued until all workers
 * have performed a restart.
 */
private global bool g_fwRPC_core_pendingRestart;


private global anytype g_fwRPC_core_msgClientCtx = "";


//------------------------------------------------------------------------------


/**
 * Initialize and run the core module.
 * There should only be one instance of the core running per project.
 *
 * If startingWorkers and maxWorkers are set to values other than -1, then the
 * core will automatically balance the current number of worker scrip WCCOActrl
 * managers running on the current project to the bounds specified.
 *
 * If distributed is set to true, the core will listen for messages from clients
 * in connected projects through the fwRPC_broker module.
 *
 * The msgClient name is automatically set to a default constant, however this
 * may be overwritten for testing purposes.
 *
 *
 * @param startingWorkers           (IN, int)    Number of workers to initialize with
 * @param maxWorkers                (IN, int)    Max number of workers at any given time
 * @param minClientContextPoolSize  (IN, int)    If the value is > 0, it ensures the min. size of client context pool
 * @param distributed               (IN, bool)   Whether or not this core is part of a non
 *                                               local RPC system.
 * @param coreName                  (IN, string) Manual specification of core name
 *
 * @returns nothing
 *
 * @throws  on PMON error
 *          if clientName is empty
 */
void fwRPC_core_initialize(
  int     startingWorkers = -1,
  int     maxWorkers = -1,
  int     minClientContextPoolSize = -1,
  bool    distributed = false,
  string  coreName = ""
) {

  if(minClientContextPoolSize > 0)
  {
    // @see FWCORE-3313
    fwRPC_client_ensureMinimumPoolSize((uint)minClientContextPoolSize);
  }

  string startupMsg = "";
  int err;
  g_fwRPC_core_distributed = distributed;

  dynClear(g_fwRPC_core_requestQueue);
  mappingClear(g_fwRPC_core_requests);
  g_fwRPC_core_last_queued_since_restart = "";

  fwRPC_functionManager_load();

  fwRPC_workerManager_initialize(
    startingWorkers,
    maxWorkers,
    "_fwRPC_core_workerDeathCB"
  );

  startupMsg = "fwRPC core started";


  if (distributed) {
    mapping msgDelegates;
    msgDelegates[fwRPC_DISTMSG_TYPE_REQUEST]     = "_fwRPC_core_hdl_c_request";
    msgDelegates[fwRPC_DISTMSG_TYPE_GET_RESULT]  = "_fwRPC_core_hdl_c_getResult";
    msgDelegates[fwRPC_DISTMSG_TYPE_ACKNOWLEDGE] = "_fwRPC_core_hdl_c_acknowledge";
    msgDelegates[fwRPC_DISTMSG_TYPE_CANCEL]      = "_fwRPC_core_hdl_c_cancel";
    msgDelegates[fwRPC_DISTMSG_TYPE_RECOVER]     = "_fwRPC_core_hdl_c_recover";
    msgDelegates[fwRPC_DISTMSG_TYPE_REGISTER]    = "_fwRPC_core_hdl_c_listSvc";

    if (coreName == "") coreName = _fwRPC_CORE_DEFAULT_NAME;

    err = fwRPC_msgClient_initialize(
      g_fwRPC_core_msgClientCtx,
      !distributed,
      fwRPC_DIST_TAG_CORE,
      coreName,
      msgDelegates
    );

    if (err < 0) {
      fwException_throw(
        fwRPC_EX_CORE_INIT,
        "[core] Error initialising msgClient: " + err
      );
    }

    startupMsg += " in distributed mode (" + coreName + ")";
  }

  fwRPC_dbg_hook(coreName, fwRPC_DBG_MODULE_CORE, "_fwRPC_core_dbg_reportCB");

  fwRPC_msgBroker_initialize();

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_STATUS,
    "_fwRPC_core_hdlStatus",
    fwRPC_DIST_TAG_WORKER
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_REQUEST,
    "_fwRPC_core_hdlRequest",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_GET_RESULT,
    "_fwRPC_core_hdlGetResult",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_SET_RESULT,
    "_fwRPC_core_hdlResult",
    fwRPC_DIST_TAG_WORKER
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_LISTSVC,
    "_fwRPC_core_hdlListSvc",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_ACKNOWLEDGE,
    "_fwRPC_core_hdlAcknowledge",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_RECOVER,
    "_fwRPC_core_hdlRecover",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_CANCEL,
    "_fwRPC_core_hdlCancel",
    fwRPC_DIST_TAG_CLIENT
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_WORKER_HEARTBEAT,
    "_fwRPC_core_hdlWorkerHeartbeat",
    fwRPC_DIST_TAG_WORKER
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_WORKER_STATE,
    "_fwRPC_core_hdlWorkerState",
    fwRPC_DIST_TAG_WORKER
  );

  fwRPC_msgBroker_accept(
    fwRPC_DISTMSG_TYPE_DISCONNECT,
    "_fwRPC_core_hdlWorkerDisconnect",
    fwRPC_DIST_TAG_WORKER
  );

  //start listening and open communication
  err = fwRPC_msgBroker_listen(true, "_fwRPC_core_hdlNewClient");
  if (err < 0) {
    fwException_throw(
      fwRPC_EX_CORE_INIT,
      "[core] Error initialising msgBroker: " + err
    );
  }

  //TODO: extract
  string dp = fwRPC_DBG_DP + ".restartCore";
  dpConnect("_fwRPC_core_restartRequest_cb", false, dp);

  fwRPC_dbg_log(fwRPC_DBG_EVENT_GENERIC, startupMsg);
}


/**
 * @internal
 *
 * Handle function definition restart requests. When triggered, the core will
 * attempt to restart all active workers in order to reload procedure
 * definitions.
 *
 * Once triggered, the core and all workers will enter a pending restart state,
 * in which all new requests are enqueued. Note that the core itself will not
 * restart, so all pending requests are kept until all workers have restarted.
 *
 * Workers will only be restarted once their state is idle, in which the workers
 * will have no active or enqueued requests.
 *
 * Once all workers have been restarted, the core will reload RPC procedure
 * definitions and disable the pending restart state, dispatching any enqueued
 * requests and accepting new requests as usual.
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_restartRequest_cb(string dp, bool restart) {
  if (restart && !g_fwRPC_core_pendingRestart) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_GENERIC,
      "BEGIN core restart"
    );

    bool noLegacyQueue = g_fwRPC_core_last_queued_since_restart == "";

    g_fwRPC_core_pendingRestart = fwRPC_workerManager_pendWorkerRestart(
      noLegacyQueue
    );

    if (!g_fwRPC_core_pendingRestart) fwRPC_functionManager_load();
  }
}


/**
 * @internal
 *
 * Work in progress -- not for release
 *
 * Part of the fwRPC debugging interface.
 *
 * This function is executed when the registered module (using fwRPC_dbg_hook())
 * receives a trigger for a status dump.
 */
public void _fwRPC_core_dbg_reportCB() {
  string name = g_fwRPC_core_msgClientCtx;
  if (name != "") name = fwRPC_msgClient_getName(g_fwRPC_core_msgClientCtx);

  dyn_string state = makeDynString(
    fwRPC_common_cleanProjectName(getSystemName()),
    name,
    (string) mappinglen(g_fwRPC_core_requests),
    (string) dynlen(g_fwRPC_core_requestQueue),
    (string) fwRPC_workerManager_getWorkerCount()
  );
  
  fwRPC_dbg_log(fwRPC_DBG_EVENT_DUMP, state);
}


/**
 * Perform a clean up before stopping the core.
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_core_shutdown() {
  fwRPC_msgBroker_stop();

  if (g_fwRPC_core_distributed) fwRPC_msgClient_stop(g_fwRPC_core_msgClientCtx);
  fwRPC_dbg_log(fwRPC_DBG_EVENT_GENERIC, "shutdown complete");
}


/**
 * @internal
 *
 * Handle worker death event. Signalled by worker manager.
 *
 * @param workerId (IN, int) dead worker
 
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_workerDeathCB(int workerId) {
  fwRPC_workerManager_deleteWorker(workerId);
  _fwRPC_core_resetRequests(workerId);
  fwRPC_dbg_log(fwRPC_DBG_EVENT_GENERIC, "dropped worker: " + workerId);
}


/**
 * @internal
 *
 * Clear the pending request queue.
 *
 * Currently used for unit testing only.
 */
void _fwRPC_core_clearQueue() {
  dynClear(g_fwRPC_core_requestQueue);
}


/**
 * @internal
 *
 * Called when msgBroker detects a new fwRPC msgClient.
 * Sends CONNECT message to newly connected client.
 *
 * @TODO: This currently does nothing at all for the client
 *
 * @param project (IN, string) source project
 * @param name    (IN, stirng) source name
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlNewClient(string project, string name) {
  if (!fwRPC_dist_isClient(name)) return;

  int err = fwRPC_msgBroker_writeMessage(
    project, name,
    fwRPC_DISTMSG_TYPE_CONNECT
  );

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Could not send CONNECT msg to " + project + " " + name
    );

    return;
  }
}


/**
 * @internal
 e
 * Handle status messages from the worker. Passes them on to the original
 * caller.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlStatus(string project, string name, anytype msg)
synchronized(g_fwRPC_core_requests) {
  anytype hdl = msg[fwRPC_DISTMSG_STATUS_HDL];
  int err;
  dyn_mixed request;

  if (!mappingHasKey(g_fwRPC_core_requests, hdl)) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "Status received for unrecognized request: " + hdl
    );
    return;
  }

  request = g_fwRPC_core_requests[hdl];
  request[fwRPC_REQUEST_MODEL_STATUS] = msg[fwRPC_DISTMSG_STATUS_STATUS];
  g_fwRPC_core_requests[hdl] = request;

  err = _fwRPC_core_sendStatusUpdate(request);

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "error sending status update for " + hdl
    );
  }

  if (fwRPC_dbg_enabled()) {
    mapping changes;
    changes["type"]   = "request";
    changes["hdl"]    = hdl;
    changes["status"] = msg[fwRPC_DISTMSG_STATUS_STATUS];
    fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
  }
}


//------------------------------------------------------------------------------
// START client/broker wrapper funcs
//------------------------------------------------------------------------------

public void _fwRPC_core_hdl_c_request(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_core_hdlRequest(project, name, msg);
}

public void _fwRPC_core_hdl_c_getResult(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_core_hdlGetResult(project, name, msg);
}

public void _fwRPC_core_hdl_c_acknowledge(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_core_hdlAcknowledge(project, name, msg);
}

public void _fwRPC_core_hdl_c_cancel(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_core_hdlCancel(project, name, msg);
}

public void _fwRPC_core_hdl_c_recover(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_core_hdlRecover(project, name, msg);
}

public void _fwRPC_core_hdl_c_listSvc(
  anytype ctx,
  string project,
  string name,
  anytype msg
) {
  _fwRPC_core_hdlListSvc(project, name, msg);
}

//------------------------------------------------------------------------------
// END client/broker wrapper funcs
//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Handle request messages from the client.
 *
 * New requests are assigned to a worker or enqueued. The immediate result of
 * which is sent back to the requesting client.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlRequest(
  string project,
  string name,
  anytype msg
) synchronized(g_fwRPC_core_requests) {
  dyn_mixed request;
  int err;

  string reqHdl = msg[fwRPC_DISTMSG_REQUEST_HDL];
  string func = msg[fwRPC_DISTMSG_REQUEST_FUNC];

  if (mappingHasKey(g_fwRPC_core_requests, reqHdl)) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_WARNING, reqHdl + " already being processed");
    return;
  }

  fwRPC_msgBroker_registerHandle(reqHdl, project, name);

  request = fwRPC_requestModel_new(
    reqHdl,
    msg[fwRPC_DISTMSG_REQUEST_PARAMS],
    msg[fwRPC_DISTMSG_REQUEST_FUNC]
  );

  err = _fwRPC_core_dispatchToWorkerOrEnqueue(request);
  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "error assigning request for " + reqHdl + " client will be notified"
    );
  }

  g_fwRPC_core_requests[reqHdl] = request;

  err = _fwRPC_core_sendStatusUpdate(request);

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "error sending request update for " + reqHdl
    );
  }

  // debug instr.
  if (fwRPC_dbg_enabled()) {
    _fwRPC_core_dbg_reportCB();

    string _name = g_fwRPC_core_msgClientCtx;
    if (_name != "") _name = fwRPC_msgClient_getName(g_fwRPC_core_msgClientCtx);

    mapping changes;
    changes["type"]   = "request";
    changes["hdl"]    = request[fwRPC_REQUEST_MODEL_ID];
    changes["func"]   = request[fwRPC_REQUEST_MODEL_FUNCTION];
    changes["params"] = request[fwRPC_REQUEST_MODEL_PARAMS];
    changes["worker"] = request[fwRPC_REQUEST_MODEL_WORKER];
    changes["core"]   = fwRPC_common_cleanProjectName(getSystemName()) + ": " + 
                        _name;
    changes["client"] = name;
    changes["status"] = request[fwRPC_REQUEST_MODEL_STATUS];
    changes["result"] = request[fwRPC_REQUEST_MODEL_RETVAL]; //in case of err
    fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
  }
}


/**
 *  @internal
 *
 *  Handle get result requests. Sends the result of a procedure call back to the
 *  client if it's state is DONE or ERROR.
 *
 *  @param project  (IN, string)  originating project
 *  @param name     (IN, string)  originating name
 *  @param msg      (IN, anytype) message buffer
 *
 *  @returns nothing
 *
 *  @throws nothing
 */
public void _fwRPC_core_hdlGetResult(
  string senderProject, string senderName,
  anytype msg
) {
  anytype hdl = msg;
  int err;
  dyn_mixed response;

  if (!mappingHasKey(g_fwRPC_core_requests, hdl)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "request " + hdl + " not registered, " + senderName + " will be notified"
    );

    response[fwRPC_DISTMSG_RESULT_HDL] = hdl;
    response[fwRPC_DISTMSG_RESULT_STATUS] = fwRPC_REQUEST_STATUS_ERROR;
    response[fwRPC_DISTMSG_RESULT_OUTPUT] = makeError(
      getErrorCatalog(fwRPC_EX_INVALID_PARAM),
      getErrorPriority(fwRPC_EX_INVALID_PARAM),
      getErrorType(fwRPC_EX_INVALID_PARAM),
      getErrorCode(fwRPC_EX_INVALID_PARAM),
      "Unable to retrieve result for request, hdl not registered" + hdl
    );

    if (fwRPC_dist_isClient(senderName)) {
      err = fwRPC_msgBroker_writeMessage(
        senderProject, senderName,
        fwRPC_DISTMSG_TYPE_SET_RESULT, response
      );
    } else if (fwRPC_dist_isBroker(senderName)) {
      err = fwRPC_msgClient_sendRequest(
        g_fwRPC_core_msgClientCtx,
        fwRPC_DISTMSG_TYPE_SET_RESULT,
        response
      );
    }

    if (err < 0) {
      fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
        "error sending result message to " + senderProject + " " + senderName
      );
    }

    return;
  }

  dyn_mixed request = g_fwRPC_core_requests[hdl];
  int status = request[fwRPC_REQUEST_MODEL_STATUS];

  response[fwRPC_DISTMSG_RESULT_HDL]    = hdl;
  response[fwRPC_DISTMSG_RESULT_STATUS] = request[fwRPC_REQUEST_MODEL_STATUS];
  response[fwRPC_DISTMSG_RESULT_OUTPUT] = request[fwRPC_REQUEST_MODEL_RETVAL];

  if (fwRPC_dist_isClient(senderName)) {
    err = fwRPC_msgBroker_writeMessage(
      senderProject, senderName,
      fwRPC_DISTMSG_TYPE_SET_RESULT, response
    );
  } else if (fwRPC_dist_isBroker(senderName)) {
    err = fwRPC_msgClient_sendRequest(
      g_fwRPC_core_msgClientCtx,
      fwRPC_DISTMSG_TYPE_SET_RESULT,
      response
    );
  }

  if (err < 0) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR, 
      "error sending result message to " + senderProject + " " + senderName
    );
  }
}


/**
 * @internal
 *
 * Handle result messages from workers. Updates local state and updates the
 * client on the request's current state.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlResult(
  string senderProject, string senderName,
  anytype msg
) synchronized(g_fwRPC_core_requests) {
  anytype hdl;
  dyn_mixed request;
  int err;

  hdl = msg[fwRPC_DISTMSG_RESULT_HDL];

  if (!mappingHasKey(g_fwRPC_core_requests, hdl)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Result received for unrecognized request: " + hdl
    );

    return;
  }

  request = g_fwRPC_core_requests[hdl];

  if (fwRPC_requestModel_hasTerminated(request)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "result received for terminated request: " + hdl + " status: " +
      request[fwRPC_REQUEST_MODEL_STATUS]
    );

    return;
  }

  request[fwRPC_REQUEST_MODEL_STATUS] = msg[fwRPC_DISTMSG_RESULT_STATUS];
  request[fwRPC_REQUEST_MODEL_RETVAL] = msg[fwRPC_DISTMSG_RESULT_OUTPUT];
  g_fwRPC_core_requests[hdl] = request;

  err = _fwRPC_core_sendStatusUpdate(request);

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "error sending status update message for request: " + hdl + " .. " + err
    );
  }

  if (fwRPC_dbg_enabled()) {
    mapping changes;
    changes["type"]   = "request";
    changes["hdl"]    = request[fwRPC_REQUEST_MODEL_ID];
    changes["result"] = request[fwRPC_REQUEST_MODEL_RETVAL];
    changes["status"] = request[fwRPC_REQUEST_MODEL_STATUS];
    fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
  }
}


/**
 * @internal
 *
 * Handle service listing messages.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlListSvc(string project, string name, anytype msg) {
  int err;
  dyn_string funcs = fwRPC_functionManager_getFunctionNames();

  if (fwRPC_dist_isClient(name)) {
    err = fwRPC_msgBroker_writeMessage(
      project, name,
      fwRPC_DISTMSG_TYPE_LISTSVC, funcs
    );
  } else if (fwRPC_dist_isBroker(name)) {
    err = fwRPC_msgClient_sendRequest(
      g_fwRPC_core_msgClientCtx,
      fwRPC_DISTMSG_TYPE_REGISTER,
      funcs
    );
  }

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "error sending listsvc message to " + project + " " + name
    );
  }
}


/**
 * @internal
 *
 * Handle acknowledge messages from the client. Ackowledges the request
 * signalling that all references to a given request can be safely deleted.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlAcknowledge(string project, string name, anytype msg)
synchronized(g_fwRPC_core_requests) {
  if (!mappingHasKey(g_fwRPC_core_requests, msg)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "request not mapped: " + msg
    );

    return;
  }

  dyn_mixed request = fwRPC_requestModel_new(msg);

  mappingRemove(g_fwRPC_core_requests, msg);

  fwRPC_msgBroker_unregisterHandle(msg);

  if (fwRPC_dbg_enabled()) {
    _fwRPC_core_dbg_reportCB();

    mapping changes;
    changes["type"]   = "request";
    changes["hdl"]    = request[fwRPC_REQUEST_MODEL_ID];
    changes["status"] = fwRPC_REQUEST_STATUS_ACKED;
    fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
  }
}


/**
 * @internal
 *
 * Handle cancel messages from the client.
 * Cancels an active request.
 * Client is not notified in case of error.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlCancel(string project, string name, anytype msg)
synchronized(g_fwRPC_core_requests) {
  if (!mappingHasKey(g_fwRPC_core_requests, msg)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_WARNING,
      "Received cancellation request for unrecognized request: " + msg
    );
    return;
  }

  dyn_mixed request = g_fwRPC_core_requests[msg];
  if (request[fwRPC_REQUEST_MODEL_STATUS] == fwRPC_REQUEST_STATUS_CANCEL) return;

  // we don't set the request state to cancel here, instead we wait for the
  // worker to send us the new state when it has finished cancellation.

  string workerProject, workerName;
  int err;

  err = fwRPC_msgBroker_getHandleCallee(msg, workerProject, workerName);
  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Could not retrieve callee for " + msg
    );
    return;
  }

  err = fwRPC_msgBroker_writeMessage(
    workerProject, workerName,
    fwRPC_DISTMSG_TYPE_CANCEL, msg
  );

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Could not send cancellation message for " + request[fwRPC_REQUEST_MODEL_ID]
    );
  }
}


/**
 * @internal
 *
 * Handle recovery messages from the client.
 * Returns all active requests assigned by the requestor.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 * 
 * @returns nothing
 *
 * @throws nothing
 *
 * @pre message caller is of type CLIENT or BROKER
 */
public void _fwRPC_core_hdlRecover(string project, string name, anytype msg)
synchronized(g_fwRPC_core_requests) {
  string cb;
  dyn_mixed response;
  anytype hdl;
  dyn_mixed request;

  if (fwRPC_dist_isClient(name)) {
    dyn_string calls;
    fwRPC_msgBroker_getHandlesCalledBy(project, name, calls);
    if (dynlen(calls) == 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_GENERIC,
        "Client requested recovery but no calls are registered for " + project +
        " " + name
      );
      return;
    }

    int callsToRecover;

    cb = msg;

    callsToRecover = dynlen(calls);
    for (int i = 1; i <= callsToRecover; ++i) {
      string workerProject, workerName;
      hdl = calls[i];
      request = g_fwRPC_core_requests[hdl];

      response[fwRPC_DISTMSG_STATUS_HDL]      = hdl;
      response[fwRPC_DISTMSG_STATUS_STATUS]   = request[fwRPC_REQUEST_MODEL_STATUS];
      response[fwRPC_DISTMSG_STATUS_CALLBACK] = cb;

      fwRPC_msgBroker_writeMessage(
        project, name,
        fwRPC_DISTMSG_TYPE_STATUS, response
      );
    }
  } else if (fwRPC_dist_isBroker(name)) {
    hdl = msg[fwRPC_DISTMSG_RECOVER_HDL];
    cb = msg[fwRPC_DISTMSG_RECOVER_CALLBACK];

    if (!mappingHasKey(g_fwRPC_core_requests, hdl)) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_WARNING,
        "Request was never registered: " + hdl
      );
      return;
    }

    request = g_fwRPC_core_requests[hdl];

    response[fwRPC_DISTMSG_STATUS_HDL]      = request[fwRPC_REQUEST_MODEL_ID];
    response[fwRPC_DISTMSG_STATUS_STATUS]   = request[fwRPC_REQUEST_MODEL_STATUS];
    response[fwRPC_DISTMSG_STATUS_CALLBACK] = cb;

    fwRPC_msgClient_send(
      g_fwRPC_core_msgClientCtx,
      fwRPC_DISTMSG_TYPE_STATUS,
      project,
      name,
      response
    );
  }
}


/**
 * @internal
 *
 * Handle worker heartbeat messages.
 * Notifies worker manager of heartbeat event.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlWorkerHeartbeat(
  string  project,
  string  name,
  anytype msg
) {
  int workerId;
  int err;
  
  workerId = (int) fwRPC_dist_getIdFromName(name);
  err = fwRPC_workerManager_notifyHeartbeat(workerId);

  if (err < 0) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_WARNING,
      "Worker " + workerId +
      " not registered but heartbeat was received. Appending as new worker."
    );

    fwRPC_msgBroker_writeMessage(
      project, name, fwRPC_DISTMSG_TYPE_WORKER_STATE
    );
  }
}


/**
 * @internal
 *
 * Handle worker state messages.
 * Notifies worker manager of new worker state.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffer
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlWorkerState(
  string  project,
  string  name,
  anytype msg
) {
  dyn_mixed worker = fwRPC_workerModel_new(
    msg[fwRPC_DISTMSG_WORKER_STATE_ID],
    msg[fwRPC_DISTMSG_WORKER_STATE_STATUS],
    msg[fwRPC_DISTMSG_WORKER_STATE_JOBS],
    msg[fwRPC_DISTMSG_WORKER_STATE_LOCKED]
  );

  fwRPC_workerManager_updateState(worker);

  if (
    g_fwRPC_core_last_queued_since_restart == "" && 
    g_fwRPC_core_pendingRestart
  ) {
    g_fwRPC_core_pendingRestart = _fwRPC_core_processPendingRestart(worker);
  }

  if (!msg[fwRPC_DISTMSG_WORKER_STATE_LOCKED]) {
    if (
      g_fwRPC_core_last_queued_since_restart == "" &&
      g_fwRPC_core_pendingRestart
    ) {
      return; //waiting for workers to restart
    }

    _fwRPC_core_processQueue();
  }
}


/**
 * @internal
 *
 * Process the current state of the local RPC system on a worker update event.
 *
 * Restarts the given worker if it is pending restart and if its state is set to
 * READY.
 *
 * If all workers have been restarted (not pending), then the procedure
 * definitions are reloaded from their DPs.
 *
 * @param updatedWorker (IN, dyn_mixed) worker object
 *
 * @returns the new pending restart state (true/false)
 *
 * @throws nothing
 */
private synchronized bool _fwRPC_core_processPendingRestart(dyn_mixed updatedWorker) {
  //check pending restart again now that we are in a sync block
  if (!g_fwRPC_core_pendingRestart) return false;

  if (
    fwRPC_workerManager_isWorkerPendingRestart(updatedWorker) &&
    fwRPC_workerModel_isIdle(updatedWorker)
  ) {
    fwRPC_workerManager_restartWorker(updatedWorker);
  }

  if (!fwRPC_workerManager_isAnyWorkerPendingRestart()) {
    fwRPC_functionManager_load();

    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_GENERIC,
      "All workers have restarted. Procedure definitions reloaded."
    );

    return false;
  }

  //still pending restart
  return true;
}


/**
 * @internal
 *
 * Handle worker disconnect message.
 *
 * @param project (IN, string)  originating project
 * @param name    (IN, string)  originating name
 * @param msg     (IN, anytype) message buffee
 *
 * @returns nothing
 *
 * @throws nothing
 */
public void _fwRPC_core_hdlWorkerDisconnect(
  string project,
  string name,
  anytype msg
) {
  fwRPC_workerManager_deleteWorker(fwRPC_dist_getIdFromName(name));
}


//------------------------------------------------------------------------------
// End remote/local message hdls
//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Resets all requests linked to a worker which are returned by the queryFormat
 *  param, if any.
 *
 * @param workerid (IN, int) The id of the worker.
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_core_resetRequests(int workerId)
synchronized(g_fwRPC_core_requests) {
  fwRPC_dbg_log(fwRPC_DBG_EVENT_GENERIC, "Resetting requests for worker " + workerId);

  int requestsToReset = mappinglen(g_fwRPC_core_requests);
  for (int i = 1; i <= requestsToReset; ++i) {
    dyn_mixed request = mappingGetValue(g_fwRPC_core_requests, i);
    anytype reqHdl;
    int err;

    if (request[fwRPC_REQUEST_MODEL_WORKER] != workerId) continue;
    if (fwRPC_requestModel_reset(request) < 0) continue; //request already finished

    reqHdl = request[fwRPC_REQUEST_MODEL_ID];

    err = _fwRPC_core_dispatchToWorkerOrEnqueue(request);
    if (err < 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "error assigning request for " + reqHdl + " client will be notified."
      );
    }

    g_fwRPC_core_requests[reqHdl] = request;

    err = _fwRPC_core_sendStatusUpdate(request);
    if (err < 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "Error sending status update for " + reqHdl
      );
    }

    if (fwRPC_dbg_enabled()) {
      _fwRPC_core_dbg_reportCB();

      mapping changes;
      changes["type"]   = "request";
      changes["hdl"]    = request[fwRPC_REQUEST_MODEL_ID];
      changes["worker"] = request[fwRPC_REQUEST_MODEL_WORKER];
      changes["status"] = request[fwRPC_REQUEST_MODEL_STATUS];
      changes["result"] = request[fwRPC_REQUEST_MODEL_RETVAL]; //in case of err
      fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
    }
  }
}


/**
 * @internal
 *
 * Return whether or not a request requires exclusive execution.
 *
 * @param request (IN, dyn_mixed) the request to check
 *
 * @return true  if the request requires exclusive execution
 *         false otherwise
 *
 * @throws nothing
 */
private bool _fwRPC_core_requestNeedsExclusive(dyn_mixed request) {
  string func = request[fwRPC_REQUEST_MODEL_FUNCTION];
  dyn_mixed functionObject;
  fwRPC_functionManager_getFunctionObject(func, functionObject);

  return functionObject[fwRPC_FUNCTION_MODEL_EXCLUSIVE];
}


/**
 * @internal
 *
 * Finds the next free worker and assigns the request to it. If no free worker
 *  exists, or if the request queue has > 0 elements, then the request will be
 *  appended to the queue.
 *
 * Request must not have already been assigned to a worker
 *
 * Updates the request status.
 *
 * @param request (IN, dyn_mixed) request to assign
 *
 * @returns fwRPC_CORE_E_SUCCESS  on success
 *          fwRPC_CORE_E_ERROR    otherwise
 *
 * @throws nothing
 */
private int _fwRPC_core_dispatchToWorkerOrEnqueue(dyn_mixed &request) {
  if (request[fwRPC_REQUEST_MODEL_WORKER] != "-1") return fwRPC_CORE_E_ERROR;
  int err;

  if (g_fwRPC_core_pendingRestart) {
    _fwRPC_core_enqueueRequest(request);
    return fwRPC_CORE_E_SUCCESS;
  }

  err = _fwRPC_core_validateRequestedFunction(request);
  if (err < 0) return fwRPC_CORE_E_ERROR;

  bool exclusive = _fwRPC_core_requestNeedsExclusive(request);

  dyn_mixed worker;
  bool exhausted = (dynlen(g_fwRPC_core_requestQueue) > 0) ||
                   (fwRPC_workerManager_findSuitableWorker(exclusive, worker) < 0);

  if (exhausted) {
    _fwRPC_core_enqueueRequest(request);
  } else {
    err = _fwRPC_core_assignRequest(worker, request);
    if (err < 0) return fwRPC_CORE_E_ERROR;
  }

  return fwRPC_CORE_E_SUCCESS;
}


/**
 * Check if the requested function is registered, and if not, set the request's
 * status and result value to a suitable error object.
 *
 * @param request (IN|OUT, dyn_mixed) request object
 *
 * @returns fwRPC_CORE_E_SUCCESS if the requested function is registered
 *          fwRPC_CORE_E_ERROR   otherwise
 *
 * @throws nothing
 */
private int _fwRPC_core_validateRequestedFunction(dyn_mixed &request) {
  string funcName = request[fwRPC_REQUEST_MODEL_FUNCTION];

  if (!fwRPC_functionManager_isRegistered(funcName)) {
    request[fwRPC_REQUEST_MODEL_STATUS] = fwRPC_REQUEST_STATUS_ERROR;
    request[fwRPC_REQUEST_MODEL_RETVAL] = makeError(
      getErrorCatalog(fwRPC_EX_INVALID_PARAM),
      getErrorPriority(fwRPC_EX_INVALID_PARAM),
      getErrorType(fwRPC_EX_INVALID_PARAM),
      getErrorCode(fwRPC_EX_INVALID_PARAM),
      "Called RPC function is not registered: " + funcName
    );

    return fwRPC_CORE_E_ERROR;
  }

  return fwRPC_CORE_E_SUCCESS;
}


/**
 * @internal
 *
 * Places a request at the end of a request queue and updates its status.
 *
 * @param request (IN, dyn_mixed) request to enqueue
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_core_enqueueRequest(dyn_mixed &request)
synchronized(g_fwRPC_core_requestQueue) {
  request[fwRPC_REQUEST_MODEL_STATUS] = fwRPC_REQUEST_STATUS_QUEUED;

  if (!g_fwRPC_core_pendingRestart) {
    g_fwRPC_core_last_queued_since_restart = request[fwRPC_REQUEST_MODEL_ID];
  }

  dynAppend(g_fwRPC_core_requestQueue, request);

  if (fwRPC_dbg_enabled()) {
    mapping changes;
    changes["type"]   = "request";
    changes["hdl"]    = request[fwRPC_REQUEST_MODEL_ID];
    changes["status"] = request[fwRPC_REQUEST_MODEL_STATUS];
    fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
  }
}


/**
 * @internal
 *
 * Dequeue requests to available workers until the queue is exhausted or no more
 *  suitable workers are available.
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_core_processQueue()
synchronized(g_fwRPC_core_requestQueue) {
  while (dynlen(g_fwRPC_core_requestQueue) > 0) {
    dyn_mixed request;
    anytype hdl;
    int err;

    hdl = g_fwRPC_core_requestQueue[1][fwRPC_REQUEST_MODEL_ID];
    request = g_fwRPC_core_requests[hdl];

    if (request[fwRPC_REQUEST_MODEL_STATUS] != fwRPC_REQUEST_STATUS_CANCEL) {
      err = _fwRPC_core_validateRequestedFunction(request);
      if (!err) {
        dyn_mixed worker;
        bool exclusive;

        exclusive = _fwRPC_core_requestNeedsExclusive(request);
        err = fwRPC_workerManager_findSuitableWorker(exclusive, worker);
        if (err < 0) {
          break; //we have run out of suitable workers
        };

        err = _fwRPC_core_assignRequest(worker, request);
        if (err != fwRPC_CORE_E_SUCCESS){
          fwRPC_dbg_log(
            fwRPC_DBG_EVENT_ERROR,
            "Could not assign worker " + worker[fwRPC_WORKER_MODEL_ID] +
            " to request " + hdl + " aborting queue processing.."
          );

          return;
        }
      }
    }

    dynRemove(g_fwRPC_core_requestQueue, 1);
    g_fwRPC_core_requests[hdl] = request;

    if (g_fwRPC_core_last_queued_since_restart == hdl) {
      g_fwRPC_core_last_queued_since_restart = "";
    }
    
    err = _fwRPC_core_sendStatusUpdate(request);
    if (err < 0) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "error sending status update for " + hdl
      );
    }

    if (fwRPC_dbg_enabled()) {
      _fwRPC_core_dbg_reportCB();

      mapping changes;
      changes["type"]   = "request";
      changes["hdl"]    = request[fwRPC_REQUEST_MODEL_ID];
      changes["worker"] = request[fwRPC_REQUEST_MODEL_WORKER];
      changes["status"] = request[fwRPC_REQUEST_MODEL_STATUS];
      changes["result"] = request[fwRPC_REQUEST_MODEL_RETVAL]; //in case of err
      fwRPC_dbg_log(fwRPC_DBG_EVENT_CHANGE, changes);
    }
  }
}


/**
 * @internal
 *
 * Delegate a request to a worker for execution.
 *
 * @param worker  (IN, dyn_mixed) worker to assign the request to
 * @param request (IN, dyn_mixed) request to assign
 *
 * @returns  0 if success
 *          -1 if communication failed
 *
 * @throws if the call has already been assigned to a worker
 */
private int _fwRPC_core_assignRequest(
  dyn_mixed worker,
  dyn_mixed &request
) {
  string workerProject, workerName;
  anytype hdl;
  dyn_mixed msg;
  int err;

  workerProject = fwRPC_common_cleanProjectName(getSystemName());
  workerName = fwRPC_DIST_TAG_LOCAL +
               "_" + fwRPC_DIST_TAG_WORKER +
               "_" + worker[fwRPC_WORKER_MODEL_ID];

  hdl = request[fwRPC_REQUEST_MODEL_ID];

  fwRPC_msgBroker_setHandleCallee(
    hdl,
    workerProject, workerName
  );

  msg[fwRPC_DISTMSG_REQUEST_HDL]    = hdl;
  msg[fwRPC_DISTMSG_REQUEST_FUNC]   = request[fwRPC_REQUEST_MODEL_FUNCTION];
  msg[fwRPC_DISTMSG_REQUEST_PARAMS] = request[fwRPC_REQUEST_MODEL_PARAMS];

  err = fwRPC_msgBroker_writeMessage(
    workerProject, workerName,
    fwRPC_DISTMSG_TYPE_REQUEST, msg
  );

  if (err) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Could not write request message to " + workerProject + " " + workerName
    );
    return fwRPC_CORE_E_ERROR;
  }

  // XXX: hack... set fake status here so that we don't accidentally affect the
  //      worker while it has been assigned something. Status will be
  //      overwritten with real data from worker itself later on.
  worker[fwRPC_WORKER_MODEL_STATUS] = fwRPC_WORKER_STATUS_PENDINGRUN;
  fwRPC_workerManager_updateState(worker);
  //end hack

  request[fwRPC_REQUEST_MODEL_WORKER] = worker[fwRPC_WORKER_MODEL_ID];
  request[fwRPC_REQUEST_MODEL_STATUS] = fwRPC_REQUEST_STATUS_ASSIGNED;

  return fwRPC_CORE_E_SUCCESS;
}


/**
 * @internal
 *
 * Send a status update message to the original caller of a request.
 *
 * @param request (IN, dyn_mixed) the request
 *
 * @returns fwRPC_CORE_E_SUCCESS      on success
 *          fwRPC_CORE_E_REQ_NOT_FOUND if the caller could not be retrieved
 *          fwRPC_CORE_E_ERROR if the message could not be sent
 *
 * @throws nothing
 */
private int _fwRPC_core_sendStatusUpdate(dyn_mixed request) {
  string callerProject, callerName;
  int err;

  err = fwRPC_msgBroker_getHandleCaller(
    request[fwRPC_REQUEST_MODEL_ID],
    callerProject, callerName
  );

  if (err < 0) return fwRPC_CORE_E_REQ_NOT_FOUND;

  dyn_mixed msg;

  msg[fwRPC_DISTMSG_STATUS_HDL]    = request[fwRPC_REQUEST_MODEL_ID];
  msg[fwRPC_DISTMSG_STATUS_STATUS] = request[fwRPC_REQUEST_MODEL_STATUS];

  if (fwRPC_dist_isClient(callerName)) {
    err = fwRPC_msgBroker_writeMessage(
      callerProject, callerName,
      fwRPC_DISTMSG_TYPE_STATUS, msg
    );
  } else if (fwRPC_dist_isBroker(callerName)) {
    err = fwRPC_msgClient_sendRequest(
      g_fwRPC_core_msgClientCtx,
      fwRPC_DISTMSG_TYPE_STATUS,
      msg
    );
  }

  if (err < 0) return fwRPC_CORE_E_ERROR;

  return fwRPC_CORE_E_SUCCESS;
}
