/** @file
 *
 * This module contains functionality for debugging fwRPC. Only one isolated
 * usage of this file is permitted per module/identifier (as seen below) given
 * the use of global variables.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_common.ctl"
#uses "CtrlSerialization"


/**
 * @ingroup Constants
 * @{
 */
const string fwRPC_DBG_DP             = "_FwRPC_Diag/core";
const string fwRPC_DBG_MODULE_DPE     = "_FwRPC_Diag/core.report_";
const string fwRPC_DBG_EVENTLIST_DPE  = "_FwRPC_Diag/core.event_list";

const string fwRPC_DBG_FLAG = "fwRPC_DEBUG";


/*
 * Module types
 * @{
 */
const string fwRPC_DBG_MODULE_CORE    = "core";
const string fwRPC_DBG_MODULE_WORKER  = "worker";
const string fwRPC_DBG_MODULE_CLIENT  = "client";
const string fwRPC_DBG_MODULE_BROKER  = "broker";
/* @} */


/**
 * Debug message types
 * @{
 */
const string fwRPC_DBG_EVENT_GENERIC  = "EV_GNRC";
const string fwRPC_DBG_EVENT_DUMP     = "EV_DUMP";
const string fwRPC_DBG_EVENT_MSG      = "EV_MESG";
const string fwRPC_DBG_EVENT_REQUEST  = "EV_REQS";
const string fwRPC_DBG_EVENT_CHANGE   = "EV_CHNG";

const string fwRPC_DBG_EVENT_ERROR    = "EV_ERRO";
const string fwRPC_DBG_EVENT_WARNING  = "EV_WARN";
const string fwRPC_DBG_EVENT_TEST     = "EV_TEST";
/* @} */

private const string fwRPC_DBG_RESTART_TRIGGER_DPE = "_FwRPC_Diag/core.restartCore:_original.._value";
/* @} */


/**
 * The module type of this debug instance to be tagged in each event.
 */
private global string g_fwRPC_debug_module;

/**
 * The identifier to be tagged in each event.
 */
private global string g_fwRPC_debug_identifier;

/**
 * The debug level.
 *   1: Trigger status dump on set
 *   2; Forced debug (ignores WCCOActrl manager dbg flags (useful for tests)
 *   Any other value: debug disabled
 */
private global int    g_fwRPC_debug_level = -1;


//------------------------------------------------------------------------------


/**
 * Check whether debugging is enabled for this module.
 *
 * @returns true  if enabled
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_dbg_enabled() {
  return (g_fwRPC_debug_level == 2) || isDbgFlag(fwRPC_DBG_FLAG);
}


/**
 * Clears all events.
 *
 * TODO: do this on a per-module basis
 */
void fwRPC_dbg_clear_events() {
  dpSetWait(fwRPC_DBG_EVENTLIST_DPE, makeDynString());
}


/**
 *  Escalate the debug level for the given module.
 *
 *  Escalating the level to 1 triggers a status dump of the specified module.
 *  Escalating the level to 2 forces debugging for the specified module. (useful
 *    for debugging).
 *
 *  If the current level is already over the requested level, the level will not
 *  be set, however, the module will still be triggered to log its status.
 *
 *  @param module (IN, string) module to escalate
 *  @param level  (IN, int)    level to escalate to
 *
 *  @returns nothing
 *
 *  @throws nothing
 */
void fwRPC_dbg_escalate(string module, int level) {
  int _level = _get_dbg_level(module);

  if (_level > level) {
    _set_dbg_level(module, _level); //trigger hook anyway
  } else {
    _set_dbg_level(module, level);
  }
}


/**
 * Deescalate the debug level for the given module.
 *
 * Deescalating the level to 1 triggers a stauts dump of the specified module.
 *   Also disables forced debugging.
 * Deescalating the level to 0 disables module status logging.
 *
 * @param module (IN, string) module to deescalate
 * @param level  (IN, int)    level to deescalate to
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_dbg_deescalate(string module, int level) {
  int _level = _get_dbg_level(module);

  if (_level < level) return;
  dpSetWait(fwRPC_DBG_MODULE_DPE + module, level);
}


/**
 * Register a module. 
 *
 * This will set the identifier and module names, allowing the correct usage of
 * fwRPC_dbg_log (in place of the explicit fwRPC_dbg_log_fwRPC_debug_module). 
 *
 * A callback is also registered to enable status polling from a debugging
 * interface. When the debug level for this module is set to a value above 0,
 * the callback is called.
 *
 * Note that this should be done once per WCCOActrl manager or WCCOAui.
 *
 * @param identifier (IN, string) an identifier for the current module.
 * @param module     (IN, string) the module name
 * @param callback   (IN, string) name of function to be called in the event of a poll
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_dbg_hook(string identifier, string module, string callback) {
  g_fwRPC_debug_identifier = identifier;
  g_fwRPC_debug_module = module;

  string dp = fwRPC_DBG_MODULE_DPE + module;
  dpConnectUserData("_diag_hook_cb", callback, true, dp);
}


/**
 * Log an event.
 *
 * @param type  (IN, string)  type of event (fwRPC_DBG_EVENT_*)
 * @param value (IN, anytype) event data
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_dbg_log(string type, anytype value) {
  fwRPC_dbg_log_module(g_fwRPC_debug_module, type, value);
}


/**
 * Log an event for a specific module name. Useful for tests.
 *
 * @param module (IN, string)  module name override (fwRPC_DBG_MODULE_*)
 * @param type   (IN, string)  event type (fwRPC_DBG_EVENT_*)
 * @param value  (IN, anytype) event data
 * @param force  (IN, bool)    force logging (useful for tests)
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_dbg_log_module(
  string module,
  string type,
  anytype value,
  bool force=false
) {
  if (type == fwRPC_DBG_EVENT_ERROR) {
    DebugTN("[" + module + "] " + type + " -- " + value);
  }

  if (type == fwRPC_DBG_EVENT_MSG) {
    DebugTN("[ DEBUG ] Explicit message event reporting is disallowed.");
    return;
  }

  if (force || fwRPC_dbg_enabled()) {
    time t = getCurrentTime();

    string event = fwRPC_dbg_pack_event(
      (string) t,
      g_fwRPC_debug_identifier,
      module,
      type,
      value
    );

    dyn_string events;
    dpGet(fwRPC_DBG_EVENTLIST_DPE, events);
    dynAppend(events, event);
    dpSet(fwRPC_DBG_EVENTLIST_DPE, events);
  }
}


/**
 * Unpack event data.
 *
 * @param event      (IN, string)   event data to unpack
 * @param t          (OUT, time)    timestamp
 * @param identifier (OUT, string)  src identifier
 * @param module     (OUT, string)  src module
 * @param type       (OUT, string)  event type
 * @param entry      (OUT, anytype) entry data
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_dbg_unpack_event(
  string  event,
  time    &t,
  string  &identifier,
  string  &module,
  string  &type,
  anytype &entry
) {
  anytype deserialized;
  deserializeFromXml(event, deserialized);
  t          = deserialized[1];
  identifier = deserialized[2];
  module     = deserialized[3];
  type       = deserialized[4];
  entry      = deserialized[5];
}


/**
 * Pack event data.
 *
 * @param t      (IN, string) timestamp
 * @param id     (IN, string) src identifier
 * @param module (IN, string) src module
 * @param type   (IN, string) event type
 * @param entry  (IN, string) entry data
 *
 * @returns serialized event
 *
 * @throws nothing
 */
string fwRPC_dbg_pack_event(
  string  t,
  string  id,
  string  module,
  string  type,
  anytype entry
) {
  dyn_mixed packet = makeDynMixed(t, id, module, type, entry);

  string serialized;
  serializeToXml(packet, serialized);
  strreplace(serialized, "\n", "");

  return serialized;
}


/**
 * Initiate a core restart. Workers in the local system will be restarted as
 * soon as possible, refreshing procedure definitions and scripts for future
 * request executions.
 *
 * @returns nothing
 *
 * @throws nothing
 */
void fwRPC_dbg_restartLocalCore() {
  dpSetWait(fwRPC_DBG_RESTART_TRIGGER_DPE, true);
}


//------------------------------------------------------------------------------


/**
 *  @internal
 *
 *  Set the debug level.
 *    A debug level of 0 supresses all logs
 *    A debug level of 1 activates logs while fwRPC_DBG_FLAG is set
 *    A debug level of 2 overrides the debug flag. Useful for testing
 *
 *  @param module (IN, string) module to set
 *  @param level  (IN, int)    level to set
 *
 *  @returns nothing
 *
 *  @throws nothing
 */
private void _set_dbg_level(string module, int level) {
  dpSetWait(fwRPC_DBG_MODULE_DPE + module, level);
}


/**
 * @internal
 *
 * Retrieve the current debug level.
 *
 * @param module (IN, stirng) module to retrieve dbg level
 *
 * @returns debug level for given module
 *
 * @throws nothing
 */
private int _get_dbg_level(string module) {
  if (g_fwRPC_debug_level == -1) {
    dpGet(fwRPC_DBG_MODULE_DPE + module, g_fwRPC_debug_level);
  }

  return g_fwRPC_debug_level;
}


/**
 * @internal
 *
 * Callback for debug level changes. triggers registered hook when the debug
 * level is 2 (forced), or debugging is enabled (fwRPC_DBG_FLAG set) and the
 * level is set to 1.
 *
 * @param callback (IN, string) user hook to call
 * @param dp       (IN, string) dbg level DP
 * @param level    (IN, int)    dbg level set
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _diag_hook_cb(string callback, string dp, int level) {
  g_fwRPC_debug_level = level;

  if (fwRPC_dbg_enabled() && g_fwRPC_debug_level == 1) {
    startThread(callback);
  }
}
