/** @file
 *
 * This library contains the functionality responsible for managing RPC functions
 *  in the core server module.
 *
 * Part of the server-side RPC library.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
**/


#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_functionModel.ctl"


/**
 *  Mapping containing all registered functions:
 *  key   = function name
 *  value = function model (see fwRPC_functionModel.ctl)
 */
private global mapping g_fwRPC_functionManager_functions;

//------------------------------------------------------------------------------


/**
 * Register a function.
 *
 * @param funcObject (IN, dyn_mixed) Function object to register
 *
 * @returns void
 *
 * @throws if the given object is not a valid function object
 *         if the function is already registered
**/
void fwRPC_functionManager_register(dyn_mixed funcObject) {
  fwRPC_functionModel_checkValid(funcObject);

  string name = funcObject[fwRPC_FUNCTION_MODEL_NAME];

  if(mappingHasKey(g_fwRPC_functionManager_functions, name)) {
    fwException_throw(
      fwRPC_EX_FUNCTION_EXISTS,
      "Cannot register function \"" + name + "\", already exists."
    );
  }

  g_fwRPC_functionManager_functions[name] = funcObject;
}


/**
 * Clears the current function registry and loads functions registered in the DB.
 *
 * @returns void
 *
 * @throws nothing
**/
void fwRPC_functionManager_load() {
  dyn_dyn_anytype functions;
  dpQuery(fwRPC_FUNCTION_QUERY_UPDATE, functions);

  fwRPC_functionManager_clearRegistry();

  int funcsToRegister = dynlen(functions);
  for(int i = 2; i <= funcsToRegister; i++) {
    dyn_mixed funcObj = fwRPC_functionModel_newFromQuery(functions[i]);

    fwRPC_functionManager_register(funcObj);
  }
}


/**
 * Check if a given object is registered.
 *
 * @param functionName (IN, string) function name to check for
 *
 * @returns true if registered
 *          otherwise false
 *
 * @throws nothing
**/
bool fwRPC_functionManager_isRegistered(string functionName) {
  return mappingHasKey(g_fwRPC_functionManager_functions, functionName);
}


/**
 * Unregister all functions
 *
 * @returns void
 *
 * @throws nothing
**/
void fwRPC_functionManager_clearRegistry() {
  mappingClear(g_fwRPC_functionManager_functions);
}


/**
 * Get a registered function. If the function is not registered it will check
 *  the DB, and register it if available.
 *
 * @param functionName   (IN, string)     name of the function
 * @param functionObject (OUT, dyn_mixed) function
 *
 * @returns void
 *
 * @throws if the function is not registered
**/
void fwRPC_functionManager_getFunctionObject(
  string functionName,
  dyn_mixed &functionObject
) {
  if(!fwRPC_functionManager_isRegistered(functionName)) {
    fwException_throw(
      fwRPC_EX_FUNCTION_EXISTS,
      "Function is not registered: " + functionName
    );
  }

  functionObject = g_fwRPC_functionManager_functions[functionName];
}


/**
 * Get a list of all registered functions.
 *
 * @returns dyn string of registered functions
 *
 * @throws nothing
**/
dyn_string fwRPC_functionManager_getFunctionNames() {
  return mappingKeys(g_fwRPC_functionManager_functions);
}
