#uses "fwUnitTestComponentAsserts.ctl"

#uses "test/fwRPC/utils.ctl"

#uses "CtrlSerialization"

#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_client.ctl"
#uses "fwRPC/fwRPC_requestModel.ctl"
#uses "fwRPC/fwRPC_RingBuffer.ctl"


/**
 * For integration tests, including monitor sub/unsub, see the integration test
 *   file.
 */


// Mocks / utils
//------------------------------------------------------------------------------
private global const string _func = "func";
private global const anytype _params = "params";
private global const string _dp_func = "_FwRPC_Function/func";
private global string _clientName = "testclient";
private global const string _result = "mock";
private global string _rbDp;
private global anytype _ctx;


void _testResult(int status, string result) {
  anytype hdl;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params);
  assertEqual(fwRPC_CLIENT_E_SUCCESS, err, "error calling");

  string id, type, proj, name;
  anytype payload;
  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);

  anytype res;
  err = fwRPC_client_getResult(_ctx, hdl, res, true);
  assertEqual(fwRPC_CLIENT_E_TIMEOUT, err, "did not timeout");

  dyn_mixed resp;
  resp[fwRPC_DISTMSG_STATUS_HDL] = hdl;
  resp[fwRPC_DISTMSG_STATUS_STATUS] = status;
  _writeResponse(fwRPC_DISTMSG_TYPE_STATUS, resp);

  delay(0, 500);

  err = fwRPC_client_getResult(_ctx, hdl, res, true, 1000);
  assertEqual(fwRPC_CLIENT_E_TIMEOUT, err, "should timeout");
  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_GET_RESULT, type, "client did not request result");

  dynClear(resp);
  resp[fwRPC_DISTMSG_RESULT_HDL] = hdl;
  resp[fwRPC_DISTMSG_RESULT_STATUS] = status;
  resp[fwRPC_DISTMSG_RESULT_OUTPUT] = result;
  _writeResponse(fwRPC_DISTMSG_TYPE_SET_RESULT, resp);

  delay(0, 500);

  err = fwRPC_client_getResult(_ctx, hdl, res, true, 1000);
  assertEqual(status, err);
  assertEqual(result, res);

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, type);
}


void _writeResponse(string type, anytype payload) {
  string id = fwRPC_UUID_generateUUID4();

  string serialized;
  serializeToXml(payload, serialized);
  strreplace(serialized, "\n", "");

  string msg = id + " " + type +
                " " + fwRPC_DIST_BRKR +
                " " + fwRPC_DIST_BRKR +
                " " + serialized;

  int err = fwRPC_RingBuffer_write(_rbDp + ".in", msg);
  assertEqual(0, err, "error writing to rb");
}


// Setup
//------------------------------------------------------------------------------

void test_fwRPC_client_setupSuite() {
  _BEGIN_SUITE("client");
  fwRPC_client_startMonitor(_ctx, _clientName, false);

  _rbDp = fwRPC_RBMBX_DPT + "/" + 
          fwRPC_DIST_TAG_LOCAL + "_" +
          fwRPC_DIST_TAG_CLIENT + "_" +
          _clientName;
}

void test_fwRPC_client_teardownSuite() {
  fwRPC_client_stopMonitor(_ctx);
  _END_SUITE();
}

void test_fwRPC_client_setup() {
  _BEGIN_TEST();
  fwRPC_RingBuffer_init(_rbDp + ".in", 50);
  fwRPC_RingBuffer_init(_rbDp + ".out", 50);
}

void test_fwRPC_client_teardown() {
  _deleteFunctions();
}

void _deleteFunctions() {
  if (dpExists(_dp_func)) dpDelete(_dp_func);
}


// tests
//------------------------------------------------------------------------------
void test_fwRPC_client_createsRequest() {
  anytype hdl;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params);
  assertEqual(0, err, "error calling");

  string id, type, proj, name;
  anytype payload;
  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);

  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);
  assertEqual(3, dynlen(payload));
  assertEqual(_func, payload[fwRPC_DISTMSG_REQUEST_FUNC]);
  assertEqual(_params, payload[fwRPC_DISTMSG_REQUEST_PARAMS]);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_createsUniqueRequests() {
  anytype hdl1;
  anytype hdl2;

  fwRPC_client_call(_ctx, hdl1, _func, _params);
  fwRPC_client_call(_ctx, hdl2, _func, _params);

  assertFalse(hdl1 == hdl2);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_cancelsEligibleRequest() {
  anytype hdl;
  int status;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params);
  assertEqual(0, err, "error calling");

  string id, type, proj, name;
  anytype payload;

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);

  err = fwRPC_client_cancel(_ctx, hdl);
  assertEqual(0, err, "error cancelling");

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);

  assertEqual(fwRPC_DISTMSG_TYPE_CANCEL, type);
  assertEqual(hdl, payload);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_getStatus() {
  anytype hdl;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params);
  assertEqual(0, err, "err calling");

  string id, type, proj, name;
  anytype payload;

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);

  int status;
  status = fwRPC_client_getStatus(_ctx, hdl);
  assertEqual(fwRPC_REQUEST_STATUS_NONE, status);

  dyn_mixed resp;
  resp[fwRPC_DISTMSG_STATUS_HDL] = hdl;
  resp[fwRPC_DISTMSG_STATUS_STATUS] = fwRPC_REQUEST_STATUS_QUEUED;

  _writeResponse(fwRPC_DISTMSG_TYPE_STATUS, resp);

  delay(0,500);
  status = fwRPC_client_getStatus(_ctx, hdl);
  assertEqual(0, err);
  assertEqual(fwRPC_REQUEST_STATUS_QUEUED, status);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_delegates_status_updates() {
  anytype hdl;
  int err;

  assertEqual(0, fwRPC_client_call(_ctx, hdl, _func, _params));

  dyn_mixed msg;
  msg[fwRPC_DISTMSG_STATUS_HDL] = hdl;
  msg[fwRPC_DISTMSG_STATUS_STATUS] = fwRPC_REQUEST_STATUS_DONE;
  _writeResponse(fwRPC_DISTMSG_TYPE_STATUS, msg);

  delay(2);

  _subscribeHdl(_ctx, hdl);

  delay(0, 500);

  assertEqual(fwRPC_REQUEST_STATUS_DONE, _getLastStatus(hdl));

  msg[fwRPC_DISTMSG_STATUS_STATUS] = fwRPC_REQUEST_STATUS_READY;
  _writeResponse(fwRPC_DISTMSG_TYPE_STATUS, msg);
  delay(0, 500);

  assertEqual(fwRPC_REQUEST_STATUS_READY, _getLastStatus(hdl));
}

//------------------------------------------------------------------------------

void test_fwRPC_client_getResultAndAck() {
  _testResult(fwRPC_REQUEST_STATUS_DONE, _result);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_getErrorAndAck() {
  _testResult(fwRPC_REQUEST_STATUS_ERROR, _result);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_acknowledgesRequests() {
  anytype hdl;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params);
  assertEqual(0, err, "error calling");

  string id, type, proj, name;
  anytype payload;
  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);

  err = fwRPC_client_acknowledge(_ctx, hdl);
  assertEqual(0, err, "error acking");

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, type);
  assertEqual(hdl, payload);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_sendsRecoveryRequest() {
  anytype hdl;

  int err = fwRPC_client_recover(_ctx, "test");
  assertEqual(0, err, "error recovering");

  string id, type, proj, name;
  anytype payload;

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_RECOVER, type);
  assertEqual("test", payload);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_sync_call_times_out_locally() {
  anytype result;
  time tprev, tnow;
  
  tprev = getCurrentTime();
  fwRPC_client_callSync(_ctx, _func, _params, result, false, 1000);
  tnow = getCurrentTime();
  assertTrue(second(tnow - tprev) == 1);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_async_call_times_out_locally() {
  anytype hdl;
  int status;
  time tprev, tnow;

  tprev = getCurrentTime();
  fwRPC_client_call(_ctx, hdl, _func, _params, false, 1);
  
  status = fwRPC_client_getStatus(_ctx, hdl);
  assert(status != fwRPC_REQUEST_STATUS_TIMEOUT, "status should not be TIMEOUT");
  delay(2);
  status = fwRPC_client_getStatus(_ctx, hdl);
  assert(status == fwRPC_REQUEST_STATUS_TIMEOUT, "status should be TIMEOUT");
}

//------------------------------------------------------------------------------

void test_fwRPC_client_does_not_accept_invalid_hdls() {
  anytype invalidHandle = "hello";

  bool thrown;
  mixed reftype;

  thrown = false;
  try {
    fwRPC_client_acknowledge(_ctx, invalidHandle);
  } catch {
    thrown = true;
  }
  assert(thrown, "fwRPC_client_acknowledge should throw");

  thrown = false;
  try {
    fwRPC_client_getResult(_ctx, invalidHandle, reftype);
  } catch {
    thrown = true;
  }
  assert(thrown, "fwRPC_client_getResult should throw");

  thrown = false;
  try {
    fwRPC_client_cancel(_ctx, invalidHandle);
  } catch {
    thrown = true;
  }
  assert(thrown, "fwRPC_client_cancel should throw");

  thrown = false;
  try {
    fwRPC_client_getStatus(_ctx, invalidHandle);
  } catch {
    thrown = true;
  }
  assert(thrown, "fwRPC_client_getStatus should throw");

  thrown = false;
  try {
    fwRPC_client_subscribeCallStatus(_ctx, invalidHandle, "x");
  } catch {
    thrown = true;
  }
  assert(thrown, "fwRPC_client_subscribeCallStatus should throw");

  thrown = false;
  try {
    fwRPC_client_unsubscribeCallStatus(_ctx, invalidHandle, "x");
  } catch {
    thrown = true;
  }
  assert(thrown, "fwRPC_client_unsubscribeCallStatus should throw");
}

//------------------------------------------------------------------------------

void test_fwRPC_client_monitor_stops_msgclient_from_arbitrary_context() {
  fwRPC_client_stopMonitor(_ctx);

  fwRPC_client_startMonitor(_ctx, _clientName, false);

  execScript("main() { fwRPC_client_stopMonitor(\"" + _ctx + "\");}", makeDynString());

  delay(2);

  //reset to original state for clean suite cleanup
  fwRPC_client_startMonitor(_ctx, _clientName, false);
}

//------------------------------------------------------------------------------

void test_fwRPC_client_recovers_from_redu_switch() {
  //In the event of an active peer switch in a redundant system, the client
  //should make attempts to request any calls which have not been completed by
  //the time the switch occurred.
  
  const string active = ".Status.Active";
  const string link0  = ".PeerAlive.Link0";
  const string recovr = ".IsRecovering";

  const string r1 = "_ReduManager";
  const string r2 = "_ReduManager_2";

  //restart client monitor with redu detect forced
  fwRPC_client_stopMonitor(_ctx);
  fwRPC_client_startMonitor(_ctx, _clientName, false, true);

  //initialize simulated redu state
  dpSetWait(
    r1 + active, true,
    r1 + link0, true,
    r1 + recovr, false,

    r2 + active, false,
    r2 + link0, true,
    r2 + recovr, false
  );

  anytype hdl;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params);
  assertEqual(0, err, "error calling");

  string id, type, proj, name;
  anytype payload;
  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);

  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);
  assertEqual(3, dynlen(payload));
  assertEqual(hdl, payload[fwRPC_DISTMSG_REQUEST_HDL]);
  assertEqual(_func, payload[fwRPC_DISTMSG_REQUEST_FUNC]);
  assertEqual(_params, payload[fwRPC_DISTMSG_REQUEST_PARAMS]);

  //Since client will not receive a result, the request will remain waiting in
  //the client's call list. This is because calls are only removed when they are
  //acknowledged.

  //Simulate a redu peer switch to trigger the client to request the call again
  dpSetWait(
    r1 + active, false,
    r1 + link0, true,
    r1 + recovr, true,

    r2 + active, true,
    r2 + link0, false,
    r2 + recovr, true
  );

  delay(1);

  bool hasMsg = _getLastMessageFromRb(_rbDp, id, type, proj, name, payload, false);
  assertEqual(false, hasMsg);

  delay(1);

  dpSetWait(
    r1 + active, false,
    r1 + link0, true,
    r1 + recovr, true,

    r2 + active, true,
    r2 + link0, false,
    r2 + recovr, false
  );

  delay(1);
  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);
  assertEqual(fwRPC_DISTMSG_TYPE_REQUEST, type);
  assertEqual(3, dynlen(payload));
  assertEqual(hdl, payload[fwRPC_DISTMSG_REQUEST_HDL]);
  assertEqual(_func, payload[fwRPC_DISTMSG_REQUEST_FUNC]);
  assertEqual(_params, payload[fwRPC_DISTMSG_REQUEST_PARAMS]);

  //note that the hdl also matches, so we can be sure to receive it from
  //previous subscriptions
  
  //cleanup client state
  fwRPC_client_stopMonitor(_ctx);
  fwRPC_client_startMonitor(_ctx, _clientName, false);

  dpSetWait(
    r1 + active, false,
    r1 + link0, false,
    r1 + recovr, false,

    r2 + active, false,
    r2 + link0, false,
    r2 + recovr, false
  );
}

//------------------------------------------------------------------------------

void test_fwRPC_client_automatically_acknowledges_requests() {
  //The client may want to perform 'fire and forget' procedure calls, where we
  //may not care about the result. In this case the client may choose to
  //automatically acknowledge the call upon receipt of a completion status
  //message.

  anytype hdl;

  int err = fwRPC_client_call(_ctx, hdl, _func, _params, true);
  assertEqual(0, err, "call should succeed");

  string id, type, proj, name;
  anytype payload;

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);

  dyn_mixed resp;
  resp[fwRPC_DISTMSG_STATUS_HDL] = hdl;
  resp[fwRPC_DISTMSG_STATUS_STATUS] = fwRPC_REQUEST_STATUS_QUEUED;

  _writeResponse(fwRPC_DISTMSG_TYPE_STATUS, resp);

  delay(0, 500);
  int status = fwRPC_client_getStatus(_ctx, hdl);
  assertEqual(fwRPC_REQUEST_STATUS_QUEUED, status);


  resp[fwRPC_DISTMSG_STATUS_STATUS] = fwRPC_REQUEST_STATUS_DONE;

  _writeResponse(fwRPC_DISTMSG_TYPE_STATUS, resp);

  delay(0, 500);
  status = fwRPC_client_getStatus(_ctx, hdl);
  assertEqual(fwRPC_REQUEST_STATUS_NONE, status);

  _getLastMessageFromRb(_rbDp, id, type, proj, name, payload);

  assertEqual(fwRPC_DISTMSG_TYPE_ACKNOWLEDGE, type);
  assertEqual(hdl, payload);
}
